import os
import dill
import warnings
import random
import numpy as np
from datetime import datetime
import logging
from itertools import chain
import torch

from .utils import safe_mkdir as _safe_mkdir
from ._version_utils import deprecated_alias, warn_deprecate_class
from abc import ABC, abstractmethod


class _LoggerMixin:
    r"""A mix-in class that has a standard Python `logger`.

    :param logger: The logger or its name (str). Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, logger=None):
        if not logger:
            self.logger = logging.getLogger('root')
        elif isinstance(logger, str):
            self.logger = logging.getLogger(logger)
        else:
            self.logger = logger


class BaseCallback(ABC, _LoggerMixin):
    r"""Base class of all callbacks.
    The class should not be directly subclassed. Instead, subclass `ActionCallback` or `ConditionCallback`.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, logger=None):
        _LoggerMixin.__init__(self, logger=logger)

    @abstractmethod
    def __call__(self, solver):
        pass  # pragma: no cover


class ActionCallback(BaseCallback):
    r"""Base class of action callbacks.
    Custom callbacks that *performs an action* should subclass this class.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def conditioned_on(self, condition_callback):
        if not isinstance(condition_callback, ConditionCallback):
            raise TypeError(f'{condition_callback} is not an instance of ConditionCallback')
        return condition_callback.set_action_callback(self)


class MonitorCallback(ActionCallback):
    r"""A callback for updating the monitor plots (and optionally saving the fig to disk).

    :param monitor: The underlying monitor responsible for plotting solutions.
    :type monitor: `neurodiffeq.monitors.BaseMonitor`
    :param fig_dir: Directory for saving monitor figs; if not specified, figs will not be saved.
    :type fig_dir: str
    :param format: Format for saving figures: {'jpg', 'png' (default), ...}.
    :type format: str
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, monitor, fig_dir=None, format=None, logger=None, **kwargs):
        super(MonitorCallback, self).__init__(logger=logger)
        self.monitor = monitor
        self.fig_dir = fig_dir
        self.format = format or 'png'

        # deprecation warnings
        for kw in ['check_against_local', 'check_against']:
            if kwargs.pop(kw, None) is not None:
                warnings.warn(
                    f'`Passing {kw}` is deprecated and ignored, '
                    f'use a `PeriodLocal` or `PeriodGlobal` to control how frequently the callback is run',
                    FutureWarning,
                )
        if kwargs.pop('repaint_last', None) is not None:
            warnings.warn(
                f'`Passing repaint_last is deprecated and ignored, '
                f'Use a `OnLastLocal` callback to plot on last epoch',
                FutureWarning,
            )
        if kwargs:
            raise ValueError(f'Unknown keyword argument(s): {list(kwargs.keys())}')

        # make dir for figs
        if fig_dir:
            _safe_mkdir(fig_dir)

    def __call__(self, solver):
        self.monitor.check(
            solver.nets,
            solver.conditions,
            history=solver.metrics_history,
        )
        if self.fig_dir:
            pic_path = os.path.join(self.fig_dir, f"epoch-{solver.global_epoch}.{self.format}")
            self.monitor.fig.savefig(pic_path, bbox_inches='tight')
            self.logger.info(f'plot saved to {pic_path}')


class StopCallback(ActionCallback):
    r"""A callback that stops the training/validation process and terminates the ``solver.fit()`` call.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        This callback should always be used together with a `ConditionCallback`,
        otherwise the ``solver.fit()`` call will exit after first epoch.
    """

    def __call__(self, solver):
        solver._stop_training = True


class CheckpointCallback(ActionCallback):
    r"""A callback that saves the networks (and their weights) to the disk.

    :param ckpt_dir:
        The directory to save model checkpoints.
        If non-existent, the directory is automatically created at instantiation time.
    :type ckpt_dir: str
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        Unless the callback is called twice within the same second, new checkpoints will not overwrite existing ones.
    """

    def __init__(self, ckpt_dir, logger=None):
        super(CheckpointCallback, self).__init__(logger=logger)
        self.ckpt_dir = ckpt_dir
        _safe_mkdir(ckpt_dir)

    def __call__(self, solver):
        now = datetime.now()
        timestr = now.strftime("%Y-%m-%d_%H-%M-%S")
        fname = os.path.join(self.ckpt_dir, timestr + ".internals")
        with open(fname, 'wb') as f:
            dill.dump(solver.get_internals("all"), f)
            self.logger.info(f"Saved checkpoint to {fname} at local epoch = {solver.local_epoch} "
                             f"(global epoch = {solver.global_epoch})")


class ReportCallback(ActionCallback):
    r"""A callback that logs the training/validation information, including

    - number of batches (train/valid)
    - batch size (train/valid)
    - generator to be used (train/valid)

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __call__(self, solver):
        self.logger.info(
            f"Starting from global epoch {solver.global_epoch - 1}\n"
            f"    training with {solver.generator['train']}\n"
            f"    validating with {solver.generator['valid']}"
        )
        tb = solver.generator['train'].size
        ntb = solver.n_batches['train']
        t = tb * ntb
        vb = solver.generator['valid'].size
        nvb = solver.n_batches['valid']
        v = vb * nvb
        self.logger.info(f"train size = {tb} x {ntb} = {t}, valid_size = {vb} x {nvb} = {v}")


ReportOnFitCallback = warn_deprecate_class(ReportCallback)


class EveCallback(ActionCallback):
    r"""A callback that readjusts the number of batches for training based on latest value of a specified metric.
    The number of batches will be :math:`\displaystyle{\left(n_0 \cdot 2^k\right)}`
    or :math:`n_\mathrm{max}` (if specified), whichever is lower,
    where :math:`\displaystyle{k=\max\left(0,\left\lfloor\log_p{\frac{v}{v_0}}\right\rfloor\right)}`
    and :math:`v` is the value of the metric in the last epoch.

    :param base_value:
        Base value of the specified metric (:math:`v_0` in the above equation).
        When the metric value is higher than ``base_value``, number of batches will be :math:`n_0`.
    :type base_value: float
    :param double_at:
        The ratio at which the batch number will be doubled (:math:`p` in the above equation).
        When :math:`\displaystyle{\frac{v}{v_0}=p^k}`,
        the number of batches will be :math:`\displaystyle{\left(n_0 \cdot 2^k\right)}`.
    :type double_at: float
    :param n_0: Minimum number of batches (:math:`n_0`). Defaults to 1.
    :type n_0: int
    :param n_max: Maximum number of batches (:math:`n_\mathrm{max}`). Defaults to infinity.
    :type n_max: int
    :param use_train: Whether to use the training (instead of validation) phase value of the metric. Defaults to True.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """
    EPS = 1e-4

    def __init__(self, base_value=1.0, double_at=0.1, n_0=1, n_max=None, use_train=True, metric='loss', logger=None):
        super(EveCallback, self).__init__(logger=logger)
        self.base_value = base_value
        self.double_at = double_at
        self.n_0 = n_0
        self.n_max = n_max or np.inf
        key = 'train' if use_train else 'valid'
        self.key = f'{key}_{metric}'

    def __call__(self, solver):
        value = solver.metrics_history[self.key][-1]
        double_times = int(self.__class__.EPS + (np.log(value) - np.log(self.base_value)) / np.log(self.double_at))
        double_times = max(double_times, 0)
        solver.n_batches['train'] = min(self.n_0 * 2 ** double_times, self.n_max)


class SimpleTensorboardCallback(ActionCallback):
    r"""A callback that writes all metric values to the disk for TensorBoard to plot.
    Tensorboard must be installed for this callback to work.

    :param writer:
        The summary writer for writing values to disk.
        Defaults to a new ``SummaryWriter`` instance created with default kwargs.
    :type writer: ``torch.utils.tensorboard.SummaryWriter``
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, writer=None, logger=None):
        super(SimpleTensorboardCallback, self).__init__(logger=logger)
        if not writer:
            self.logger.info('No writer specified, creating a SummaryWriter automatically.')
        try:
            from torch.utils.tensorboard import SummaryWriter
        except ImportError as e:
            raise ImportError(f"TensorBoard doesn't seem to be installed. See the following\n{e}")

        self.writer = writer or SummaryWriter()

    def __call__(self, solver):
        for name, values in solver.metrics_history.items():
            self.writer.add_scalar(
                tag=name,
                scalar_value=values[-1] if values else np.nan,
                global_step=solver.global_epoch,
            )


class SetCriterion(ActionCallback):
    r"""A callback that sets the ``criterion`` (a.k.a. loss function) of the solver.
    Best used together with a condition callback.

    :param criterion:
        The loss function to be set for the solver. It can be

        - An instance of ``torch.nn.modules.loss._Loss``
          which computes loss of the PDE/ODE residuals against a zero tensor.
        - A callable object which maps residuals, function values, and input coordinates to a scalar loss; or
        - A str which is present in ``neurodiffeq.losses._losses.keys()``.

    :type criterion: ``torch.nn.modules.loss._Loss`` or callable or str.
    :param reset:
        If True, the criterion will be reset every time the callback is called.
        Otherwise, the criterion will only be set once.
        Defaults to False.
    :type reset: bool
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, criterion, reset=False, logger=None):
        super(SetCriterion, self).__init__(logger=logger)
        self.criterion = criterion
        self.reset = reset
        self.called = False

    def __call__(self, solver):
        if self.reset or (not self.called):
            self.called = True
            # noinspection PyProtectedMember
            solver._set_criterion(self.criterion)


class SetOptimizer(ActionCallback):
    r"""A callback that sets the optimizer of the solver. Best used together with a condition callback.

    - If an optimizer *instance* is passed, it must contain a sequence of parameters to be updated.
    - If an optimizer *subclass* is passed, optimizer_args and optimizer_kwargs can be supplied.

    :param optimizer: Optimizer instance (or its class) to be set.
    :type optimizer: type or ``torch.optim.Optimizer``
    :param optimizer_args:
        Positional arguments to be passed to the optimizer constructor in addition to the parameter sequence.
        Ignored if optimizer is an instance (instead of a class).
    :type optimizer_args: tuple
    :param optimizer_kwargs:
        Keyword arguments to be passed to the optimizer constructor in addition to the parameter sequence.
        Ignored if optimizer is an instance (instead of a class).
    :type optimizer_kwargs: dict
    :param reset:
        If True, the optimizer will be reset every time the callback is called.
        Otherwise, the optimizer will only be set once.
        Defaults to False.
    :type reset: bool
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, optimizer, optimizer_args=None, optimizer_kwargs=None, reset=False, logger=None):
        super(SetOptimizer, self).__init__(logger=logger)
        self.optimizer = optimizer
        self.optimizer_args = optimizer_args or ()
        self.optimizer_kwargs = optimizer_kwargs or {}
        self.reset = reset
        self.called = False

    def __call__(self, solver):
        if self.reset or (not self.called):
            self.called = True
            if isinstance(self.optimizer, torch.optim.Optimizer):
                solver.optimizer = self.optimizer
            elif issubclass(self.optimizer, torch.optim.Optimizer):
                params = chain.from_iterable(net.parameters() for net in solver.nets)
                solver.optimizer = self.optimizer(params, *self.optimizer_args, **self.optimizer_kwargs)
            else:
                raise TypeError(f"Unknown optimizer instance/type {self.optimizer}")


class ConditionCallback(BaseCallback):
    r"""Base class of condition callbacks.
    Custom callbacks that *determines whether some action shall be performed* should subclass this class and overwrite
    the ``.condition`` method.

    Instances of ``ConditionCallback`` (and its children classes) support (short-circuit) evaluation of
    common boolean operations: ``&`` (and), ``|`` (or), ``~`` (not), and ``^`` (xor).

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, logger=None):
        super(ConditionCallback, self).__init__(logger=logger)
        self.action_callback = None

    def set_action_callback(self, action_callback):
        if not isinstance(action_callback, ActionCallback):
            raise TypeError(f'{action_callback} is not an instance of ActionCallback')
        self.action_callback = action_callback
        return self

    @abstractmethod
    def condition(self, solver) -> bool:
        pass  # pragma: no cover

    def __call__(self, solver):
        if self.condition(solver):
            if self.action_callback:
                self.logger.debug(f"condition of {self} met, running the underlying callback {self.action_callback}")
                self.action_callback(solver)
            else:
                self.logger.warning(f"condition of {self} met, but no underlying action callback is set; skipping")
        else:
            self.logger.debug(f"condition of {self} not met")

    def __and__(self, other):
        return AndCallback(condition_callbacks=[self, other], logger=self.logger)

    def __or__(self, other):
        return OrCallback(condition_callbacks=[self, other], logger=self.logger)

    def __invert__(self):
        return NotCallback(condition_callback=self, logger=self.logger)

    def __xor__(self, other):
        return XorCallback(condition_callbacks=[self, other], logger=self.logger)


class AndCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True iff none of its sub-``ConditionCallback`` s evaluates to False.

    :param condition_callbacks: List of sub-``ConditionCallback`` s.
    :type condition_callbacks: list[``ConditionCallback``]
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        ``c = AndCallback([c1, c2, c3])`` can be simplified as ``c = c1 & c2 & c3``.
    """

    def __init__(self, condition_callbacks, logger=None):
        super(AndCallback, self).__init__(logger=logger)
        self.condition_callbacks = condition_callbacks

    def condition(self, solver) -> bool:
        for cond_cb in self.condition_callbacks:
            c = cond_cb.condition(solver)
            if not c:
                return False
        return True


class OrCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to False iff none of its sub-``ConditionCallback`` s evaluates to True.

    :param condition_callbacks: List of sub-``ConditionCallback`` s.
    :type condition_callbacks: list[``ConditionCallback``]
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        ``c = OrCallback([c1, c2, c3])`` can be simplified as ``c = c1 | c2 | c3``.
    """

    def __init__(self, condition_callbacks, logger=None):
        super(OrCallback, self).__init__(logger=logger)
        self.condition_callbacks = condition_callbacks

    def condition(self, solver) -> bool:
        for cond_cb in self.condition_callbacks:
            if cond_cb.condition(solver):
                return True
        return False


class NotCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True iff its sub-``ConditionCallback`` evaluates to False.

    :param condition_callback: The sub-``ConditionCallback`` .
    :type condition_callback: ConditionCallback
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        ``c = NotCallback(c1)`` can be simplified as ``c = ~c1``.
    """

    def __init__(self, condition_callback, logger=None):
        super(NotCallback, self).__init__(logger=logger)
        self.condition_callback = condition_callback

    def condition(self, solver) -> bool:
        return not self.condition_callback.condition(solver)


class XorCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to False iff
    evenly many of its sub-``ConditionCallback`` s evaluates to True.

    :param condition_callbacks: List of sub-``ConditionCallback`` s.
    :type condition_callbacks: list[``ConditionCallback``]
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``

    .. note::
        ``c = XorCallback([c1, c2, c3])`` can be simplified as ``c = c1 ^ c2 ^ c3``.
    """

    def __init__(self, condition_callbacks, logger=None):
        super(XorCallback, self).__init__(logger=logger)
        self.condition_callbacks = condition_callbacks

    def condition(self, solver) -> bool:
        return sum(1 for cond_cb in self.condition_callbacks if cond_cb.condition(solver)) % 2 == 1


class TrueCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which always evaluates to True.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def condition(self, solver) -> bool:
        return True


class FalseCallback(ConditionCallback):
    r"""A ``ConditionCallback`` which always evaluates to False.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def condition(self, solver) -> bool:
        return False


class OnFirstLocal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only on the first local epoch.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def condition(self, solver) -> bool:
        return solver.local_epoch == 1


class OnFirstGlobal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only on the first global epoch.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def condition(self, solver) -> bool:
        return solver.global_epoch == 1


class OnLastLocal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only on the last local epoch.

    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def condition(self, solver) -> bool:
        return solver.local_epoch == solver._max_local_epoch


class PeriodLocal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only when the local epoch count equals
    :math:`\mathrm{period}\times n + \mathrm{offset}`.

    :param period: Period of the callback.
    :type period: int
    :param offset: Offset of the period. Defaults to 0.
    :type offset: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, period, offset=0, logger=None):
        super(PeriodLocal, self).__init__(logger=logger)
        self.period = period
        self.offset = offset % period

    def condition(self, solver) -> bool:
        return solver.local_epoch % self.period == self.offset


class PeriodGlobal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only when the global epoch count equals
    :math:`\mathrm{period}\times n + \mathrm{offset}`.

    :param period: Period of the callback.
    :type period: int
    :param offset: Offset of the period. Defaults to 0.
    :type offset: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, period, offset=0, logger=None):
        super(PeriodGlobal, self).__init__(logger=logger)
        self.period = period
        self.offset = offset % period

    def condition(self, solver) -> bool:
        return solver.global_epoch % self.period == self.offset


class ClosedIntervalLocal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only when
    :math:`l_0 \leq l \leq l_1`, where :math:`l` is the local epoch count.

    :param min: Lower bound of the closed interval (:math:`l_0` in the above inequality). Defaults to None.
    :type min: int
    :param max: Upper bound of the closed interval (:math:`l_1` in the above inequality). Defaults to None.
    :type max: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, min=None, max=None, logger=None):
        super(ClosedIntervalLocal, self).__init__(logger=logger)
        self.min = - np.inf if min is None else min
        self.max = np.inf if max is None else max

    def condition(self, solver) -> bool:
        return self.min <= solver.local_epoch <= self.max


class ClosedIntervalGlobal(ConditionCallback):
    r"""A ``ConditionCallback`` which evaluates to True only when
    :math:`g_0 \leq g \leq g_1`, where :math:`g` is the global epoch count.

    :param min: Lower bound of the closed interval (:math:`g_0` in the above inequality). Defaults to None.
    :type min: int
    :param max: Upper bound of the closed interval (:math:`g_1` in the above inequality). Defaults to None.
    :type max: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, min=None, max=None, logger=None):
        super(ClosedIntervalGlobal, self).__init__(logger=logger)
        self.min = - np.inf if min is None else min
        self.max = np.inf if max is None else max

    def condition(self, solver) -> bool:
        return self.min <= solver.global_epoch <= self.max


class Random(ConditionCallback):
    r"""A ``ConditionCallback`` which has a certain probability of evaluating to True.

    :param probability: The probability of this callback evaluating to True (between 0 and 1).
    :type probability: float
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, probability, logger=None):
        super(Random, self).__init__(logger=logger)
        if probability < 0 or probability > 1:
            raise ValueError('probability must lie in [0, 1]')
        self.probability = probability

    def condition(self, solver) -> bool:
        return random.random() < self.probability


class _RepeatedMetricChange(ConditionCallback):
    def __init__(self, use_train=True, metric='loss', repetition=1, logger=None):
        super(_RepeatedMetricChange, self).__init__(logger=logger)
        key = 'train' if use_train else 'valid'
        self.key = f'{key}_{metric}'
        self.times_required = repetition
        self.so_far = 0

    @abstractmethod
    def _last_satisfied(self, last, second2last):
        return last > second2last

    def condition(self, solver) -> bool:
        history = solver.metrics_history[self.key]
        if len(history) >= 2 and self._last_satisfied(last=history[-1], second2last=history[-2]):
            self.so_far += 1
        else:
            self.so_far = 0
        return self.so_far >= self.times_required


class RepeatedMetricUp(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric for the latest :math:`n` epochs
    kept increasing by at least some margin.

    :param at_least_by: The said margin.
    :type at_least_by: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should increase by the said margin (the :math:`n`).
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, at_least_by=0.0, use_train=True, metric='loss', repetition=1, logger=None):
        super(RepeatedMetricUp, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.at_least_by = at_least_by

    def _last_satisfied(self, last, second2last):
        return last >= second2last + self.at_least_by


class RepeatedMetricDown(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric for the latest :math:`n` epochs
    kept decreasing by at least some margin.

    :param at_least_by: The said margin.
    :type at_least_by: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should decrease by the said margin (the :math:`n`).
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, at_least_by=0.0, use_train=True, metric='loss', repetition=1, logger=None):
        super(RepeatedMetricDown, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.at_least_by = at_least_by

    def _last_satisfied(self, last, second2last):
        return last <= second2last - self.at_least_by


class RepeatedMetricConverge(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric for the latest :math:`n` epochs
    kept converging within some tolerance :math:`\varepsilon`.

    :param epsilon: The said tolerance.
    :type epsilon: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should converge within said tolerance.
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, epsilon, use_train=True, metric='loss', repetition=1, logger=None):
        super(RepeatedMetricConverge, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.epsilon = abs(epsilon)

    def _last_satisfied(self, last, second2last):
        return abs(last - second2last) < self.epsilon


class RepeatedMetricDiverge(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric for the latest :math:`n` epochs
    kept diverging beyond some gap.

    :param gap: The said gap.
    :type gap: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should diverge beyond said gap.
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, gap, use_train=True, metric='loss', repetition=1, logger=None):
        super(RepeatedMetricDiverge, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.gap = abs(gap)

    def _last_satisfied(self, last, second2last):
        return abs(last - second2last) > self.gap


class RepeatedMetricBelow(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric has been less than
    a given value :math:`v` for the latest :math:`n` epochs.

    :param threshold: The value `v`.
    :type threshold: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should diverge beyond said gap.
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, threshold, use_train, metric, repetition, logger):
        super(RepeatedMetricBelow, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.threshold = threshold

    def _last_satisfied(self, last, second2last):
        return last < self.threshold


class RepeatedMetricAbove(_RepeatedMetricChange):
    r"""A ``ConditionCallback`` which evaluates to True if a certain metric has been greater than
    a given value :math:`v` for the latest :math:`n` epochs.

    :param threshold: The value `v`.
    :type threshold: float
    :param use_train: Whether to use the metric value in the training (rather than validation) phase.
    :type use_train: bool
    :param metric:
        Name of which metric to use. Must be 'loss' or present in ``solver.metrics_fn.keys()``. Defaults to 'loss'.
    :type metric: str
    :param repetition: Number of times the metric should diverge beyond said gap.
    :type repetition: int
    :param logger: The logger (or its name) to be used for this callback. Defaults to the 'root' logger.
    :type logger: str or ``logging.Logger``
    """

    def __init__(self, threshold, use_train, metric, repetition, logger):
        super(RepeatedMetricAbove, self).__init__(
            use_train=use_train, metric=metric, repetition=repetition, logger=logger
        )
        self.threshold = threshold

    def _last_satisfied(self, last, second2last):
        return last > self.threshold


class ProgressBarCallBack(ActionCallback):
    def __call__(self, solver):
        a = solver.local_epoch
        b = solver._max_local_epoch

        progress = int(a / b * 100)
        print('#' * progress + '.' * (100 - progress), end='\r', flush=True)



import pytest
import numpy as np
import torch
from neurodiffeq.networks import FCNN
from neurodiffeq.conditions import NoCondition
from neurodiffeq.operators import grad as cartesian_grad, spherical_grad, cylindrical_grad
from neurodiffeq.operators import div as cartesian_div, spherical_div, cylindrical_div
from neurodiffeq.operators import curl as cartesian_curl, spherical_curl, cylindrical_curl
from neurodiffeq.operators import laplacian as cartesian_laplacian, spherical_laplacian, cylindrical_laplacian
from neurodiffeq.operators import vector_laplacian as cartesian_vector_laplacian, spherical_vector_laplacian, \
    cylindrical_vector_laplacian

EPS = 1e-4
N_SAMPLES = 1000

CARTESIAN_X = (
    torch.rand(N_SAMPLES, 1, requires_grad=True),  # x
    torch.rand(N_SAMPLES, 1, requires_grad=True),  # y
    torch.rand(N_SAMPLES, 1, requires_grad=True),  # z
)

SPHERICAL_X = (
    torch.rand(N_SAMPLES, 1, requires_grad=True) + 0.1,  # r
    torch.rand(N_SAMPLES, 1, requires_grad=True) * (np.pi - 0.2) + 0.1,  # theta
    torch.rand(N_SAMPLES, 1, requires_grad=True) * np.pi * 2,  # phi
)

CYLINDRICAL_X = (
    torch.rand(N_SAMPLES, 1, requires_grad=True) + 0.1,  # rho
    torch.rand(N_SAMPLES, 1, requires_grad=True) * np.pi * 2,  # phi
    torch.rand(N_SAMPLES, 1, requires_grad=True),  # z
)


def vector_field(x):
    cond = NoCondition()
    return tuple(cond.enforce(FCNN(3, 1), *x) for _ in range(3))


def scalar_field(x):
    cond = NoCondition()
    return cond.enforce(FCNN(3, 1), *x)


def is_zero(t):
    if isinstance(t, (tuple, list)):
        for i in t:
            if not is_zero(i):
                return False
        return True
    elif isinstance(t, torch.Tensor):
        return t.detach().cpu().max() < EPS
    else:
        raise ValueError(f"t must be list, tuple or tensor; got {type(t)}")


@pytest.mark.parametrize(
    argnames=['x', '_curl', '_div'],
    argvalues=[
        (CARTESIAN_X, cartesian_curl, cartesian_div),
        (CYLINDRICAL_X, cylindrical_curl, cylindrical_div),
        (SPHERICAL_X, spherical_curl, spherical_div),
    ]
)
def test_div_curl(x, _curl, _div):
    U = vector_field(x)
    curl_u = _curl(*U, *x)
    div_curl_u = _div(*curl_u, *x)
    assert is_zero(div_curl_u), div_curl_u


@pytest.mark.parametrize(
    argnames=['x', '_grad', '_curl'],
    argvalues=[
        (CARTESIAN_X, cartesian_grad, cartesian_curl),
        (CYLINDRICAL_X, cylindrical_grad, cylindrical_curl),
        (SPHERICAL_X, spherical_grad, spherical_curl),
    ]
)
def test_curl_grad(x, _grad, _curl):
    u = scalar_field(x)
    grad_u = _grad(u, *x)
    curl_grad_u = _curl(*grad_u, *x)
    assert is_zero(curl_grad_u), curl_grad_u


@pytest.mark.parametrize(
    argnames=['x', '_grad', '_div', '_laplacian'],
    argvalues=[
        (CARTESIAN_X, cartesian_grad, cartesian_div, cartesian_laplacian),
        (CYLINDRICAL_X, cylindrical_grad, cylindrical_div, cylindrical_laplacian),
        (SPHERICAL_X, spherical_grad, spherical_div, spherical_laplacian),
    ]
)
def test_div_grad(x, _div, _grad, _laplacian):
    u = scalar_field(x)
    grad_u = _grad(u, *x)
    div_grad_u = _div(*grad_u, *x)
    lap_u = _laplacian(u, *x)
    delta = div_grad_u - lap_u
    assert is_zero(delta), delta


@pytest.mark.parametrize(
    argnames=['x', '_grad', '_div', '_laplacian'],
    argvalues=[
        (CARTESIAN_X, cartesian_grad, cartesian_div, cartesian_laplacian),
        (CYLINDRICAL_X, cylindrical_grad, cylindrical_div, cylindrical_laplacian),
        (SPHERICAL_X, spherical_grad, spherical_div, spherical_laplacian),
    ]
)
def test_laplacian(x, _div, _grad, _laplacian):
    test_div_grad(x, _div, _grad, _laplacian)


@pytest.mark.parametrize(
    argnames=['x', '_grad', '_div', '_curl', '_vector_laplacian'],
    argvalues=[
        (CARTESIAN_X, cartesian_grad, cartesian_div, cartesian_curl, cartesian_vector_laplacian),
        (CYLINDRICAL_X, cylindrical_grad, cylindrical_div, cylindrical_curl, cylindrical_vector_laplacian),
        (SPHERICAL_X, spherical_grad, spherical_div, spherical_curl, spherical_vector_laplacian),
    ]
)
def test_curl_curl(x, _grad, _div, _curl, _vector_laplacian):
    U = vector_field(x)
    curl_curl_u = _curl(*_curl(*U, *x), *x)
    grad_div_u = _grad(_div(*U, *x), *x)
    vec_lap_u = _vector_laplacian(*U, *x)

    vec_delta = [cc - (gd - vl) for cc, gd, vl in zip(curl_curl_u, grad_div_u, vec_lap_u)]
    assert is_zero(vec_delta), vec_delta


@pytest.mark.parametrize(
    argnames=['x', '_grad', '_div', '_curl', '_vector_laplacian'],
    argvalues=[
        (CARTESIAN_X, cartesian_grad, cartesian_div, cartesian_curl, cartesian_vector_laplacian),
        (CYLINDRICAL_X, cylindrical_grad, cylindrical_div, cylindrical_curl, cylindrical_vector_laplacian),
        (SPHERICAL_X, spherical_grad, spherical_div, spherical_curl, spherical_vector_laplacian),
    ]
)
def test_vector_laplacian(x, _grad, _div, _curl, _vector_laplacian):
    test_curl_curl(x, _grad, _div, _curl, _vector_laplacian)



import pytest
import torch
from neurodiffeq import diff
from neurodiffeq.losses import _losses as losses

N = 100


def pde_system(u, v, w, x, y):
    return [
        diff(u, x, order=2) + diff(u, y, order=2),
        diff(v, x, order=2) + diff(v, y, order=2),
        diff(w, x, order=2) + diff(w, y, order=2),
    ]


def get_rfx(n_input, n_output, n_equation):
    coords = [torch.rand((N, 1), requires_grad=True) for _ in range(n_input)]
    coords_tensor = torch.cat(coords, dim=1)
    funcs = [torch.sigmoid(torch.sum(coords_tensor, dim=1, keepdim=True)) for _ in range(n_output)]
    residual = [diff(funcs[0], coords[0]) + funcs[0] for _ in range(n_equation)]
    residual = torch.cat(residual, dim=1)
    return residual, funcs, coords


@pytest.mark.parametrize(argnames='n_input', argvalues=[1, 3])
@pytest.mark.parametrize(argnames='n_output', argvalues=[1, 3])
@pytest.mark.parametrize(argnames='n_equation', argvalues=[1, 3])
@pytest.mark.parametrize(
    argnames=('loss_name', 'loss_fn'),
    argvalues=losses.items(),
)
def test_losses(n_input, n_output, n_equation, loss_name, loss_fn):
    r, f, x = get_rfx(n_input, n_output, n_equation)
    loss = loss_fn(r, f, x)
    assert loss.shape == (), f"{loss_name} doesn't output scalar"
    assert loss.requires_grad, f"{loss_name} doesn't require gradient"



import torch
import pytest
from neurodiffeq import diff
from neurodiffeq.networks import FCNN
from neurodiffeq.operators import grad
from neurodiffeq.operators import _split_u_x
from neurodiffeq.operators import div
from neurodiffeq.operators import curl
from neurodiffeq.operators import laplacian
from neurodiffeq.operators import vector_laplacian

N_SAMPLES = 10

f1 = lambda x, y, z: x ** 2 + y ** 2 + z ** 2
g1 = lambda x, y, z: [2 * x, 2 * y, 2 * z]
f2 = lambda x, y, z: x ** 2 + y ** 2
g2 = lambda x, y, z: [2 * x, 2 * y, torch.zeros_like(z, requires_grad=z.requires_grad)]
f3 = lambda x, y, z: torch.zeros_like(x, requires_grad=True)
g3 = lambda x, y, z: list(map(lambda v: torch.zeros_like(v, requires_grad=v.requires_grad), [x, y, z]))


@pytest.fixture
def x3():
    return [torch.rand((N_SAMPLES, 1), requires_grad=True) for _ in range(3)]


@pytest.mark.parametrize(
    argnames=['u_len', 'x_len', 'err'],
    argvalues=[(0, 0, True), (3, 3, False), (3, 4, True)]
)
def test__split_u_x(u_len, x_len, err):
    us = [torch.zeros(N_SAMPLES, 1, requires_grad=True) for _ in range(u_len)]
    xs = [torch.zeros(N_SAMPLES, 1, requires_grad=True) for _ in range(x_len)]

    if not err:
        vs, ys = _split_u_x(*us, *xs)
        for u, v in zip(us, vs):
            assert u is v
        for x, y in zip(xs, ys):
            assert x is y

    else:
        with pytest.raises(RuntimeError):
            _split_u_x(*us, *xs)


@pytest.mark.parametrize(
    argnames=['f', 'g_func'],
    argvalues=list(zip([f1, f2, f3], [g1, g2, g3])),
)
def test_grad(x3, f, g_func):
    g3_true = g_func(*x3)
    g3_pred = grad(f(*x3), *x3)

    for x, g_true, g_pred in zip(x3, g3_true, g3_pred):
        assert isinstance(g_pred, torch.Tensor)
        assert g_pred.requires_grad
        assert torch.isclose(g_true, g_pred).all()


@pytest.mark.parametrize(argnames='n_dim', argvalues=[1, 2, 3])
def test_div(n_dim):
    xs = [torch.rand(N_SAMPLES, 1, requires_grad=True) for _ in range(n_dim)]
    nets = [FCNN(n_dim, 1) for _ in range(n_dim)]
    us = [net(torch.cat(xs, dim=1)) for net in nets]

    grads = [diff(u, x) for u, x in zip(us, xs)]
    ans = torch.cat(grads, dim=1).sum(dim=1, keepdim=True)

    output = div(*us, *xs)
    assert (output == ans).all()


def test_curl(x3):
    nets = [FCNN(3, 1) for _ in range(3)]
    x, y, z = x3
    u, v, w = [net(torch.cat(x3, dim=1)) for net in nets]

    ans = [
        diff(w, y) - diff(v, z),
        diff(u, z) - diff(w, x),
        diff(v, x) - diff(u, y),
    ]
    output = curl(u, v, w, x, y, z)

    for a, o in zip(ans, output):
        assert (a == o).all()


@pytest.mark.parametrize(argnames='n_dim', argvalues=[1, 2, 3])
def test_laplacian(n_dim):
    xs = [torch.rand(N_SAMPLES, 1, requires_grad=True) for _ in range(n_dim)]
    net = FCNN(n_dim, 1)
    u = net(torch.cat(xs, dim=1))
    ans = torch.cat([diff(u, x, order=2) for x in xs], dim=1).sum(dim=1, keepdim=True)
    output = laplacian(u, *xs)
    assert (ans == output).all()


def test_vector_laplacian(x3):
    nets = [FCNN(3, 1) for _ in range(3)]
    x, y, z = x3
    u, v, w = [net(torch.cat(x3, dim=1)) for net in nets]
    ans = [gd - cc for gd, cc in zip(
        grad(div(u, v, w, x, y, z), x, y, z),
        curl(*curl(u, v, w, x, y, z), x, y, z),
    )]
    output = vector_laplacian(u, v, w, x, y, z)

    for a, o in zip(ans, output):
        assert torch.isclose(o, a).all()



import pytest
import torch
import torch.nn as nn
from torch import sin, cos
import numpy as np
from neurodiffeq import diff
from neurodiffeq.generators import GeneratorSpherical
from neurodiffeq.function_basis import ZonalSphericalHarmonics
from neurodiffeq.networks import FCNN
from neurodiffeq.operators import spherical_curl
from neurodiffeq.operators import spherical_grad
from neurodiffeq.operators import spherical_div
from neurodiffeq.operators import spherical_laplacian
from neurodiffeq.operators import spherical_vector_laplacian
from neurodiffeq.operators import spherical_to_cartesian, cartesian_to_spherical


@pytest.fixture(autouse=True)
def magic():
    torch.manual_seed(42)
    np.random.seed(42)


class HarmonicsNN(nn.Module):
    def __init__(self, degrees, harmonics_fn):
        super(HarmonicsNN, self).__init__()
        self.net_r = FCNN(1, n_output_units=len(degrees))
        self.harmonics_fn = harmonics_fn

    def forward(self, r, theta, phi):
        R = self.net_r(r)
        Y = self.harmonics_fn(theta, phi)
        return (R * Y).sum(dim=1, keepdim=True)


EPS = 1e-4
degrees = list(range(10))


@pytest.fixture
def x():
    n_points, r_min, r_max = 1024, 1.0, 10.0
    g = GeneratorSpherical(n_points, r_min=r_min, r_max=r_max)
    return [t.reshape(-1, 1) for t in g.get_examples()]


@pytest.fixture
def U(x):
    F = [HarmonicsNN(degrees, ZonalSphericalHarmonics(degrees=degrees)) for _ in range(3)]
    return tuple(f(*x) for f in F)


@pytest.fixture
def u(x):
    return HarmonicsNN(degrees, ZonalSphericalHarmonics(degrees=degrees))(*x)


def test_cartesian_to_spherical():
    x = torch.rand(1000, 1, requires_grad=True)
    y = torch.rand(1000, 1, requires_grad=True)
    z = torch.rand(1000, 1, requires_grad=True)
    r, theta, phi = cartesian_to_spherical(x, y, z)
    assert torch.allclose(r * torch.sin(theta) * cos(phi), x)
    assert torch.allclose(r * torch.sin(theta) * sin(phi), y)
    assert torch.allclose(r * torch.cos(theta), z)


def test_spherical_to_cartesian():
    r = torch.rand(1000, 1, requires_grad=True)
    theta = torch.rand(1000, 1, requires_grad=True) * np.pi
    phi = torch.rand(1000, 1, requires_grad=True) * np.pi * 2
    x, y, z = spherical_to_cartesian(r, theta, phi)
    assert torch.allclose(r * torch.sin(theta) * cos(phi), x)
    assert torch.allclose(r * torch.sin(theta) * sin(phi), y)
    assert torch.allclose(r * torch.cos(theta), z)


def test_spherical_div(U, x):
    out = spherical_div(*U, *x)
    ur, utheta, uphi = U
    r, theta, phi = x
    ans = diff(r ** 2 * ur, r) / r ** 2 + \
          diff(utheta * sin(theta), theta) / (r * sin(theta)) + \
          diff(uphi, phi) / (r * sin(theta))

    assert torch.allclose(out, ans)


def test_spherical_grad(u, x):
    out_r, out_theta, out_phi = spherical_grad(u, *x)
    r, theta, phi = x
    assert torch.allclose(out_r, diff(u, r))
    assert torch.allclose(out_theta, diff(u, theta) / r)
    assert torch.allclose(out_phi, diff(u, phi) / (r * sin(theta)))


def test_spherical_curl(U, x):
    out_r, out_theta, out_phi = spherical_curl(*U, *x)
    ur, utheta, uphi = U
    r, theta, phi = x
    assert torch.allclose(out_r, (diff(uphi * sin(theta), theta) - diff(utheta, phi)) / (r * sin(theta)))
    assert torch.allclose(out_theta, (diff(ur, phi) / sin(theta) - diff(r * uphi, r)) / r)
    assert torch.allclose(out_phi, (diff(r * utheta, r) - diff(ur, theta)) / r)


def test_spherical_laplacian(u, x):
    out = spherical_laplacian(u, *x)
    r, theta, phi = x
    assert torch.allclose(
        out,
        diff(r ** 2 * diff(u, r), r) / r ** 2
        + diff(sin(theta) * diff(u, theta), theta) / (r ** 2 * sin(theta))
        + diff(u, phi, order=2) / (r ** 2 * sin(theta) ** 2)
    )


def test_spherical_vector_laplacian(U, x):
    out_r, out_theta, out_phi = spherical_vector_laplacian(*U, *x)
    ur, utheta, uphi = U
    r, theta, phi = x

    def scalar_lap(u):
        return diff(r ** 2 * diff(u, r), r) / r ** 2 \
               + diff(sin(theta) * diff(u, theta), theta) / (r ** 2 * sin(theta)) \
               + diff(u, phi, order=2) / (r ** 2 * sin(theta) ** 2)

    assert torch.allclose(
        out_r,
        scalar_lap(ur)
        - 2 * ur / r ** 2
        - 2 / (r ** 2 * sin(theta)) * diff(utheta * sin(theta), theta)
        - 2 / (r ** 2 * sin(theta)) * diff(uphi, phi)
    )
    assert torch.allclose(
        out_theta,
        scalar_lap(utheta)
        - utheta / (r ** 2 * sin(theta) ** 2)
        + 2 / r ** 2 * diff(ur, theta)
        - 2 * cos(theta) / (r ** 2 * sin(theta) ** 2) * diff(uphi, phi)
    )
    assert torch.allclose(
        out_phi,
        scalar_lap(uphi)
        - uphi / (r ** 2 * sin(theta) ** 2)
        + 2 / (r ** 2 * sin(theta)) * diff(ur, phi)
        + 2 * cos(theta) / (r ** 2 * sin(theta) ** 2) * diff(utheta, phi)
    )



from neurodiffeq.hypersolver.hypersolver import DiscreteSolution1D, Hypersolver
from neurodiffeq.hypersolver.numerical_solvers import Euler
import torch
import numpy as np


def test_discrete_solution_1d():
    try_solver = Euler()
    ts, us = try_solver.solve(lambda u, t: [-u], 1, 0, 1, 100)

    solution = DiscreteSolution1D(ts, us)

    other_ts = torch.rand(100)
    assert min(ts) < min(other_ts)
    assert max(other_ts) < max(ts)

    other_us, = solution(other_ts)
    true_us = torch.exp(-other_ts)
    assert torch.allclose(other_us, true_us, rtol=1e-2)

    ts, fs, gs = try_solver.solve(lambda u, v, t: [v, -u], [1, 1], 0, np.pi, 314)
    solution = DiscreteSolution1D(ts, fs, gs)
    other_ts = torch.rand(100) * np.pi

    other_fs, other_gs = solution(other_ts)
    fs_true = torch.sin(other_ts) + torch.cos(other_ts)
    gs_true = torch.cos(other_ts) - torch.sin(other_ts)

    assert torch.allclose(other_fs, fs_true, rtol=1e-2, atol=1e-2)
    assert torch.allclose(other_gs, gs_true, rtol=1e-2, atol=1e-2)


def test_hypersolver():
    solver = Hypersolver(
        func=lambda u, t: [-u],
        u0=1,
        t0=0,
        tn=1,
        n_steps=100,
        sol=lambda t: [torch.exp(-t)],
        numerical_solver=Euler(),
    )
    solver.fit(max_epochs=1)
    sol = solver.get_solution()
    ts = torch.rand(100)
    hypersolver_sol = sol(ts)
    analytical_sol = solver.solution(ts)

    for h_s, a_s in zip(hypersolver_sol, analytical_sol):
        assert torch.allclose(h_s, a_s, rtol=1e-2)

    solver = Hypersolver(
        func=lambda u, v, t: [v, -u],
        u0=[1, 1],
        t0=0,
        tn=np.pi,
        n_steps=314,
        sol=lambda t: [torch.sin(t) + torch.cos(t), torch.cos(t) - torch.sin(t)],
        numerical_solver=Euler(),
    )
    solver.fit(max_epochs=10000)
    sol = solver.get_solution()
    ts = torch.rand(100) * np.pi
    hypersolver_sol = sol(ts)
    analytical_sol = solver.solution(ts)

    for h_s, a_s in zip(hypersolver_sol, analytical_sol):
        assert torch.allclose(h_s, a_s, rtol=1e-2, atol=1e-2)



# Consider these convention:
# For user facing part, use 1-d tensor (when representing a collection of 2-d points, use 2 1-d tensors).
# For non-user facing part, avoid reshaping and use 1-d tensor as much as possible.
# In function signatures, let u comes before x, let x comes before t
# Use x and t for distinguish values for x or t; Use xx and tt when corresponding entries of xx and tt are
# supposed to be paired to represent a point. ([xx, tt] is often the Cartesian product of x and t)
from abc import ABC, abstractmethod
import torch
import numpy as np
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.tri as tri
from copy import deepcopy

# return the Cartesian product of x and t.
def _cartesian_prod_dims(x, t, x_grad=True, t_grad=True):
    xt = torch.cartesian_prod(x, t)
    xx = torch.squeeze(xt[:, 0])
    xx.requires_grad = x_grad
    tt = torch.squeeze(xt[:, 1])
    tt.requires_grad = t_grad
    return xx, tt

class Approximator(ABC):
    r"""The base class of approximators. An approximator is an approximation of the differential equation's solution.
    It knows the parameters in the neural network, and how to calculate the loss function and the metrics.
    """
    @abstractmethod
    def __call__(self):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def parameters(self):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def calculate_loss(self):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def calculate_metrics(self):
        raise NotImplementedError  # pragma: no cover


class SingleNetworkApproximator1DSpatialTemporal(Approximator):
    r"""An approximator to approximate the solution of a 1D time-dependent problem.
    The boundary condition will be enforced by a regularization term in the loss function
    and the initial condition will be enforced by transforming the output of the
    neural network.

    :param single_network: A neural network with 2 input nodes (x, t) and 1 output node.
    :type single_network: `torch.nn.Module`
    :param pde: The PDE to solve. If the PDE is :math:`F(u, x, t) = 0` then `pde`
        should be a function that maps :math:`(u, x, t)` to :math:`F(u, x, t)`.
    :type pde: function
    :param initial_condition: A first order initial condition.
    :type initial_condition: `temporal.FirstOrderInitialCondition`
    :param boundary_conditions: A list of boundary conditions.
    :type boundary_conditions: list[`temporal.BoundaryCondition`]
    :param boundary_strictness: The regularization parameter, defaults to 1.
        a larger regularization parameter enforces the boundary conditions more strictly.
    :type boundary_strictness: float
    """
    def __init__(self, single_network, pde, initial_condition, boundary_conditions, boundary_strictness=1.):
        self.single_network = single_network
        self.pde = pde
        self.initial_condition = initial_condition
        self.boundary_conditions = boundary_conditions
        self.boundary_strictness = boundary_strictness

    def __call__(self, xx, tt):
        xx = torch.unsqueeze(xx, dim=1)
        tt = torch.unsqueeze(tt, dim=1)
        xt = torch.cat((xx, tt), dim=1)
        uu = torch.exp(-tt) * self.initial_condition.u0(xx) + (1 - torch.exp(-tt)) * self.single_network(xt)
        return torch.squeeze(uu)

    def parameters(self):
        return self.single_network.parameters()

    def calculate_loss(self, xx, tt, x, t):
        uu = self.__call__(xx, tt)

        equation_mse = torch.mean(self.pde(uu, xx, tt)**2)

        boundary_mse = self.boundary_strictness * sum(self._boundary_mse(t, bc) for bc in self.boundary_conditions)

        return equation_mse + boundary_mse

    def _boundary_mse(self, t, bc):
        x = next(bc.points_generator)

        xx, tt = _cartesian_prod_dims(x, t, x_grad=True, t_grad=False)
        uu = self.__call__(xx, tt)
        return torch.mean(bc.form(uu, xx, tt)**2)

    def calculate_metrics(self, xx, tt, x, t, metrics):
        uu = self.__call__(xx, tt)

        return {
            metric_name: metric_func(uu, xx, tt)
            for metric_name, metric_func in metrics.items()
        }


class SingleNetworkApproximator2DSpatial(Approximator):
    r"""An approximator to approximate the solution of a 2D steady-state problem.
    The boundary condition will be enforced by a regularization term in the loss function.

    :param single_network: A neural network with 2 input nodes (x, y) and 1 output node.
    :type single_network: `torch.nn.Module`
    :param pde: The PDE to solve. If the PDE is :math:`F(u, x, y) = 0` then `pde`
        should be a function that maps :math:`(u, x, y)` to :math:`F(u, x, y)`.
    :type pde: function
    :param boundary_conditions: A list of boundary conditions.
    :type boundary_conditions: list[`temporal.BoundaryCondition`]
    :param boundary_strictness: The regularization parameter, defaults to 1.
        A larger regularization parameter enforces the boundary conditions more strictly.
    :type boundary_strictness: float
    """
    def __init__(self, single_network, pde, boundary_conditions, boundary_strictness=1.):
        self.single_network = single_network
        self.pde = pde
        self.boundary_conditions = boundary_conditions
        self.boundary_strictness = boundary_strictness

    def __call__(self, xx, yy):
        xx = torch.unsqueeze(xx, dim=1)
        yy = torch.unsqueeze(yy, dim=1)
        xy = torch.cat((xx, yy), dim=1)
        uu = self.single_network(xy)
        return torch.squeeze(uu)

    def parameters(self):
        return self.single_network.parameters()

    def calculate_loss(self, xx, yy):
        uu = self.__call__(xx, yy)

        equation_mse = torch.mean(self.pde(uu, xx, yy)**2)

        boundary_mse = self.boundary_strictness * sum(self._boundary_mse(bc) for bc in self.boundary_conditions)

        return equation_mse + boundary_mse

    def _boundary_mse(self, bc):
        xx, yy = next(bc.points_generator)
        uu = self.__call__(xx, yy)
        return torch.mean(bc.form(uu, xx, yy) ** 2)

    def calculate_metrics(self, xx, yy, metrics):
        uu = self.__call__(xx, yy)

        return {
            metric_name: metric_func(uu, xx, yy)
            for metric_name, metric_func in metrics.items()
        }


class SingleNetworkApproximator2DSpatialSystem(Approximator):
    r"""An approximator to approximate the solution of a 2D steady-state differential equation system.
    The boundary condition will be enforced by a regularization term in the loss function.

    :param single_network:
        A neural network with 2 input nodes (x, y) and n output node (n is the number of
        dependent variables in the differential equation system)
    :type single_network: `torch.nn.Module`
    :param pde:
        The PDE system to solve. If the PDE is :math:`F_i(u_1, u_2, ..., u_n, x, y) = 0`
        where :math:`u_i` is the i-th dependent variable,
        then `pde` should be a function that maps :math:`(u_1, u_2, ..., u_n, x, y)` to
        a list where the i-th entry is :math:`F_i(u_1, u_2, ..., u_n, x, y)`.
    :type pde: callable
    :param boundary_conditions:
        A list of boundary conditions
    :type boundary_conditions: list[`temporal.BoundaryCondition`]
    :param boundary_strictness:
        The regularization parameter, defaults to 1.
        a larger regularization parameter enforces the boundary conditions more strictly.
    :type boundary_strictness: float
    """
    def __init__(self, single_network, pde, boundary_conditions, boundary_strictness=1.):
        self.single_network = single_network
        self.pde = pde
        self.boundary_conditions = boundary_conditions
        self.boundary_strictness = boundary_strictness

    def __call__(self, xx, yy):
        xx = torch.unsqueeze(xx, dim=1)
        yy = torch.unsqueeze(yy, dim=1)
        xy = torch.cat((xx, yy), dim=1)
        uu = self.single_network(xy)
        return tuple(torch.squeeze(uu[:, i]) for i in range(uu.shape[1]))

    def parameters(self):
        return self.single_network.parameters()

    def calculate_loss(self, xx, yy):
        uu = self.__call__(xx, yy)

        equation_mse = sum(
            torch.mean(eq**2)
            for eq in self.pde(*uu, xx, yy)
        )

        boundary_mse = self.boundary_strictness * sum(self._boundary_mse(bc) for bc in self.boundary_conditions)

        return equation_mse + boundary_mse

    def _boundary_mse(self, bc):
        xx, yy = next(bc.points_generator)
        uu = self.__call__(xx, yy)
        return torch.mean(bc.form(*uu, xx, yy) ** 2)

    def calculate_metrics(self, xx, yy, metrics):
        uu = self.__call__(xx, yy)

        return {
            metric_name: metric_func(*uu, xx, yy)
            for metric_name, metric_func in metrics.items()
        }


class SingleNetworkApproximator2DSpatialTemporal(Approximator):
    r"""An approximator to approximate the solution of a 2D time-dependent problem.
    The boundary condition will be enforced by a regularization term in the loss function
    and the initial condition will be enforced by transforming the output of the neural network.

    :param single_network:
        A neural network with 3 input nodes (x, y, t) and 1 output node.
    :type single_network: `torch.nn.Module`
    :param pde:
        The PDE system to solve. If the PDE is :math:`F(u, x, y, t) = 0`
        then `pde` should be a function that maps :math:`(u, x, y, t)` to :math:`F(u, x, y, t)`.
    :type pde: callable
    :param initial_condition:
        A first order initial condition.
    :type initial_condition: `temporal.FirstOrderInitialCondition` or `temporal.SecondOrderInitialCondition`
    :param boundary_conditions:
        A list of boundary conditions.
    :type boundary_conditions: list[`temporal.BoundaryCondition`]
    :param boundary_strictness:
        The regularization parameter, defaults to 1.
        a larger regularization parameter enforces the boundary conditions more strictly.
    :type boundary_strictness: float
    """
    def __init__(self, single_network, pde, initial_condition, boundary_conditions, boundary_strictness=1.):
        self.single_network = single_network
        self.pde = pde
        self.u0 = initial_condition.u0
        self.u0dot = initial_condition.u0dot if hasattr(initial_condition, 'u0dot') else None
        self.boundary_conditions = boundary_conditions
        self.boundary_strictness = boundary_strictness

    def __call__(self, xx, yy, tt):
        xx = torch.unsqueeze(xx, dim=1)
        yy = torch.unsqueeze(yy, dim=1)
        tt = torch.unsqueeze(tt, dim=1)
        xyt = torch.cat((xx, yy, tt), dim=1)
        if self.u0dot is None:
            uu = torch.exp(-tt) * self.u0(xx, yy) + (1 - torch.exp(-tt)) * self.single_network(xyt)
        else:
            # not sure about this line
            uu = (1 - (1 - torch.exp(-tt))**2) * self.u0(xx, yy) \
                 + (1 - torch.exp(-tt)) * self.u0dot(xx, yy) \
                 + (1 - torch.exp(-tt))**2 * self.single_network(xyt)
        return torch.squeeze(uu)

    def parameters(self):
        return self.single_network.parameters()

    def calculate_loss(self, xx, yy, tt, x, y, t):
        uu = self.__call__(xx, yy, tt)

        equation_mse = torch.mean(self.pde(uu, xx, yy, tt)**2)

        boundary_mse = self.boundary_strictness * sum(self._boundary_mse(t, bc) for bc in self.boundary_conditions)

        return equation_mse + boundary_mse

    def _boundary_mse(self, t, bc):
        x, y = next(bc.points_generator)

        xx, tt = _cartesian_prod_dims(x, t, x_grad=True, t_grad=False)
        yy, tt = _cartesian_prod_dims(y, t, x_grad=True, t_grad=False)
        uu = self.__call__(xx, yy, tt)
        return torch.mean(bc.form(uu, xx, yy, tt) ** 2)

    def calculate_metrics(self, xx, yy, tt, x, y, t, metrics):
        uu = self.__call__(xx, yy, tt)

        return {
            metric_name: metric_func(uu, xx, yy, tt)
            for metric_name, metric_func in metrics.items()
        }


class FirstOrderInitialCondition:
    r"""A first order initial condition. It is used to initialize ``temporal.Approximator``\s.

    :param u0:
        A function representing the initial condition.
        If we are solving for :math:`u`, then `u0` is :math:`u\bigg|_{t=0}`.
        The input of the function depends on where it is used.

        - If it is used as the input for `temporal.SingleNetworkApproximator1DSpatialTemporal`,
          then `u0` should map :math:`x` to :math:`u(x, t)\bigg|_{t = 0}`.
        - If it is used as the input for `temporal.SingleNetworkApproximator2DSpatialTemporal`,
          then `u0` should map :math:`(x, y)` to :math:`u(x, y, t)\bigg|_{t = 0}`.
    :type u0: callable
    """
    def __init__(self, u0):
        self.u0 = u0


class SecondOrderInitialCondition:
    r"""A second order initial condition. It is used to initialize ``temporal.Approximator``\s.

    :param u0:
        A function representing the initial condition.
        If we are solving for is :math:`u`, then ``u0`` is :math:`u\bigg|_{t=0}`.
        The input of the function dependes on where it is used.

        - If it is used as the input for ``temporal.SingleNetworkApproximator1DSpatialTemporal``,
          then ``u0`` should map :math:`x` to :math:`u(x, t)\bigg|_{t = 0}`.
        - If it is used as the input for ``temporal.SingleNetworkApproximator2DSpatialTemporal``,
          then ``u0`` should map :math:`(x, y)` to :math:`u(x, y, t)\bigg|_{t = 0}`.
    :type u0: callable
    :param u0dot:
        A function representing the initial derivative w.r.t. time.
        If we are solving for is :math:`u`, then ``u0dot`` is :math:`\dfrac{\partial u}{\partial t}\bigg|_{t=0}`.
        The input of the function depends on where it is used.

        - If it is used as the input for ``temporal.SingleNetworkApproximator1DSpatialTemporal``,
          then ``u0`` should map :math:`x` to :math:`\dfrac{\partial u}{\partial t}\bigg|_{t = 0}`.
        - If it is used as the input for `temporal.SingleNetworkApproximator2DSpatialTemporal`,
          then ``u0`` should map :math:`(x, y)` to :math:`\dfrac{\partial u}{\partial t}\bigg|_{t = 0}`.
    :type u0dot: callable
    """
    def __init__(self, u0, u0dot):
        self.u0 = u0
        self.u0dot = u0dot


class BoundaryCondition:
    r"""A boundary condition. It is used to initialize ``temporal.Approximator``\s.

    :param form: The form of the boundary condition.

        - For a 1D time-dependent problem, if the boundary condition demands that :math:`B(u, x) = 0`,
          then ``form`` should be a function that maps :math:`u, x, t` to :math:`B(u, x)`.
        - For a 2D steady-state problem, if the boundary condition demands that :math:`B(u, x, y) = 0`,
          then ``form`` should be a function that maps :math:`u, x, y` to :math:`B(u, x, y)`.
        - For a 2D steady-state system, if the boundary condition demands that :math:`B(u_i, x, y) = 0`,
          then ``form`` should be a function that maps :math:`u_1, u_2, ..., u_n, x, y` to :math:`B(u_i, x, y)`.
        - For a 2D time-dependent problem, if the boundary condition demands that :math:`B(u, x, y) = 0`,
          then ``form`` should be a function that maps :math:`u, x, y, t` to :math:`B(u_i, x, y)`.

        Basically the function signature of ``form`` should be
        the same as the ``pde`` function of the given ``temporal.Approximator``.
    :type form: callable
    :param points_generator:
        A generator that generates points on the boundary.
        It can be a `temporal.generator_1dspatial`, `temporal.generator_2dspatial_segment`,
        or a generator written by user.
    :type points_genrator: generator
    """
    def __init__(self, form, points_generator):
        self.form = form
        self.points_generator = points_generator


def generator_1dspatial(size, x_min, x_max, random=True):
    r"""Return a generator that generates 1D points range from x_min to x_max

    :param size:
        Number of points to generated when ``__next__`` is invoked.
    :type size: int
    :param x_min:
        Lower bound of x.
    :type x_min: float
    :param x_max:
        Upper bound of x.
    :type x_max: float
    :param random:
        - If set to False, then return equally spaced points range from ``x_min`` to ``x_max``.
        - If set to True then generate points randomly.

        Defaults to True.
    :type random: bool
    """
    seg_len = (x_max-x_min) / size
    linspace_lo = x_min + seg_len*0.5
    linspace_hi = x_max - seg_len*0.5
    center = torch.linspace(linspace_lo, linspace_hi, size)
    noise_lo = -seg_len*0.5
    while True:
        if random:
            noise = seg_len*torch.rand(size) + noise_lo
            yield center + noise
        else:
            yield center


def generator_2dspatial_segment(size, start, end, random=True):
    r"""Return a generator that generates 2D points in a line segment.

    :param size:
        Number of points to generated when `__next__` is invoked.
    :type size: int
    :param x_min:
        Lower bound of x.
    :type x_min: float
    :param x_max:
        Upper bound of x.
    :type x_max: float
    :param y_min:
        Lower bound of y.
    :type y_min: float
    :param y_max:
        Upper bound of y.
    :type y_max: float
    :param random:

        - If set to False, then return a grid where the points are equally spaced in the x and y dimension.
        - If set to True then generate points randomly.

        Defaults to True.
    :type random: bool
    """
    x1, y1 = start
    x2, y2 = end
    step = 1./size
    center = torch.linspace(0. + 0.5*step, 1. - 0.5*step, size)
    noise_lo = -step*0.5
    while True:
        if random:
            noise = step*torch.rand(size) + noise_lo
            center = center + noise
        yield x1 + (x2-x1)*center, y1 + (y2-y1)*center


def generator_2dspatial_rectangle(size, x_min, x_max, y_min, y_max, random=True):
    r"""Return a generator that generates 2D points in a rectangle.

    :param size:
        Number of points to generated when `__next__` is invoked.
    :type size: int
    :param start:
        The starting point of the line segment.
    :type start: tuple[float, float]
    :param end:
        The ending point of the line segment.
    :type end: tuple[float, float]
    :param random:
        - If set to False, then return eqally spaced points range from `start` to `end`.
        - If set to Rrue then generate points randomly.

        Defaults to True.
    :type random: bool
    """
    x_size, y_size = size
    x_generator = generator_1dspatial(x_size, x_min, x_max, random)
    y_generator = generator_1dspatial(y_size, y_min, y_max, random)
    while True:
        x = next(x_generator)
        y = next(y_generator)
        xy = torch.cartesian_prod(x, y)
        xx = torch.squeeze(xy[:, 0])
        yy = torch.squeeze(xy[:, 1])
        yield xx, yy


def generator_temporal(size, t_min, t_max, random=True):
    r"""Return a generator that generates 1D points range from t_min to t_max

    :param size:
        Number of points to generated when `__next__` is invoked.
    :type size: int
    :param t_min:
        Lower bound of t.
    :type t_min: float
    :param t_max:
        Upper bound of t.
    :type t_max: float
    :param random:
        - If set to False, then return eqally spaced points range from t_min to t_max.
        - If set to True then generate points randomly.

        Defaults to True
    :type random: bool
    """
    seg_len = (t_max - t_min) / size
    linspace_lo = t_min + seg_len * 0.5
    linspace_hi = t_max - seg_len * 0.5
    center = torch.linspace(linspace_lo, linspace_hi, size)
    noise_lo = -seg_len * 0.5
    while True:
        if random:
            noise = seg_len * torch.rand(size) + noise_lo
            yield center + noise
        else:
            yield center


class MonitorMinimal:
    r"""A monitor that shows the loss function and custom metrics.
    """
    def __init__(self, check_every):
        self.using_non_gui_backend = matplotlib.get_backend() == 'agg'
        self.check_every = check_every

        self.fig = plt.figure(figsize=(20, 8))
        self.ax1 = self.fig.add_subplot(121)
        self.ax2 = self.fig.add_subplot(122)

    def check(self, approximator, history):

        self.ax1.clear()
        self.ax1.plot(history['train_loss'], label='training loss')
        self.ax1.plot(history['valid_loss'], label='validation loss')
        self.ax1.set_title('loss during training')
        self.ax1.set_ylabel('loss')
        self.ax1.set_xlabel('epochs')
        self.ax1.set_yscale('log')
        self.ax1.legend()

        self.ax2.clear()
        for metric_name, metric_values in history.items():
            if metric_name == 'train_loss' or metric_name == 'valid_loss':
                continue
            self.ax2.plot(metric_values, label=metric_name)
        self.ax2.set_title('metrics during training')
        self.ax2.set_ylabel('metrics')
        self.ax2.set_xlabel('epochs')
        self.ax2.set_yscale('log')
        # if there's not custom metrics, then there won't be any labels in this axis
        if len(history) > 2:
            self.ax2.legend()

        self.fig.canvas.draw()
        if not self.using_non_gui_backend:
            plt.pause(0.05)   # pragma: no cover (we are not using gui backend for testing)


class Monitor1DSpatialTemporal:
    r"""A monitor for 1D time-dependent problems.
    """
    def __init__(self, check_on_x, check_on_t, check_every):
        self.using_non_gui_backend = matplotlib.get_backend() == 'agg'

        self.xx_tensor, self.tt_tensor = _cartesian_prod_dims(check_on_x, check_on_t)
        self.x_array = check_on_x.clone().detach().cpu().numpy()
        self.t_array = check_on_t.clone().detach().cpu().numpy()
        self.check_every = check_every
        self.t_color = torch.linspace(0, 1, len(check_on_t)).detach().cpu().numpy()

        self.fig = plt.figure(figsize=(30, 8))
        self.ax1 = self.fig.add_subplot(131)
        self.ax2 = self.fig.add_subplot(132)
        self.ax3 = self.fig.add_subplot(133)

    def check(self, approximator, history):
        uu_array = approximator(self.xx_tensor, self.tt_tensor).detach().cpu().numpy()

        self.ax1.clear()
        for i, t_c in enumerate(zip(self.t_array, self.t_color)):
            u_t = uu_array[i::len(self.t_array)]
            t, c = t_c
            t = float(t)
            c = cm.viridis(c)

            self.ax1.plot(self.x_array, u_t, color=c, label=f't = {t:.2E}')
        self.ax1.legend()
        self.ax1.set_title('approximation')

        self.ax2.clear()
        self.ax2.plot(history['train_loss'], label='training loss')
        self.ax2.plot(history['valid_loss'], label='validation loss')
        self.ax2.set_title('loss during training')
        self.ax2.set_ylabel('loss')
        self.ax2.set_xlabel('epochs')
        self.ax2.set_yscale('log')
        self.ax2.legend()

        self.ax3.clear()
        for metric_name, metric_values in history.items():
            if metric_name == 'train_loss' or metric_name == 'valid_loss':
                continue
            self.ax3.plot(metric_values, label=metric_name)
        self.ax3.set_title('metrics during training')
        self.ax3.set_ylabel('metrics')
        self.ax3.set_xlabel('epochs')
        self.ax3.set_yscale('log')
        # if there's not custom metrics, then there won't be any labels in this axis
        if len(history) > 2:
            self.ax3.legend()

        self.fig.canvas.draw()
        if not self.using_non_gui_backend:
            plt.pause(0.05)   # pragma: no cover (we are not using gui backend for testing)


class Monitor2DSpatialTemporal:
    r"""A monitor for 2D time-dependent problems.
    """
    def __init__(self, check_on_x, check_on_y, check_on_t, check_every):
        self.using_non_gui_backend = matplotlib.get_backend() == 'agg'

        xy_tensor = torch.cartesian_prod(check_on_x, check_on_y)
        self.xx_tensor = torch.squeeze(xy_tensor[:, 0])
        self.yy_tensor = torch.squeeze(xy_tensor[:, 1])
        self.tt_tensors = [
            torch.ones(len(xy_tensor)) * t
            for t in check_on_t
        ]
        self.xx_array = self.xx_tensor.clone().detach().cpu().numpy()
        self.yy_array = self.yy_tensor.clone().detach().cpu().numpy()
        self.t_array = check_on_t.clone().detach().cpu().numpy()
        self.check_every = check_every

        self.fig = None
        self.axs = []  # subplots
        self.cbs = []  # color bars

    @staticmethod
    def _create_contour(ax, xx, yy, uu):
        triang = tri.Triangulation(xx, yy)
        contour = ax.tricontourf(triang, uu, cmap='coolwarm')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal', adjustable='box')
        return contour

    def check(self, approximator, history):
        if not self.fig:
            # initialize the figure and axes here so that the Monitor knows the number of dependent variables and
            # size of the figure, number of the subplots, etc.
            # one for each time slice, plus one for training and validation loss, plus one for metrics
            n_axs = len(self.t_array)+2
            n_row, n_col = (n_axs+1) // 2, 2
            self.fig = plt.figure(figsize=(20, 8*n_row))
            for i in range(n_axs):
                self.axs.append(self.fig.add_subplot(n_row, n_col, i+1))
            for i in range(n_axs-2):
                self.cbs.append(None)

        for i, ax in enumerate(self.axs[:-2]):
            ax.clear()
            uu_array = approximator(self.xx_tensor, self.yy_tensor, self.tt_tensors[i]).detach().cpu().numpy()
            cs = self._create_contour(ax, self.xx_array, self.yy_array, uu_array)
            if self.cbs[i] is None:
                self.cbs[i] = self.fig.colorbar(cs, format='%.0e', ax=ax)
            else:
                self.cbs[i].mappable.set_clim(vmin=uu_array.min(), vmax=uu_array.max())
            ax.set_title(f'approximation t = {self.t_array[i]:.2E}')

        self.axs[-2].clear()
        self.axs[-2].plot(history['train_loss'], label='training loss')
        self.axs[-2].plot(history['valid_loss'], label='validation loss')
        self.axs[-2].set_title('loss during training')
        self.axs[-2].set_ylabel('loss')
        self.axs[-2].set_xlabel('epochs')
        self.axs[-2].set_yscale('log')
        self.axs[-2].legend()

        self.axs[-1].clear()
        for metric_name, metric_values in history.items():
            if metric_name == 'train_loss' or metric_name == 'valid_loss':
                continue
            self.axs[-1].plot(metric_values, label=metric_name)
        self.axs[-1].set_title('metrics during training')
        self.axs[-1].set_ylabel('metrics')
        self.axs[-1].set_xlabel('epochs')
        self.axs[-1].set_yscale('log')

        # if there's not custom metrics, then there won't be any labels in this axis
        if len(history) > 2:
            self.axs[-1].legend()

        self.fig.canvas.draw()
        if not self.using_non_gui_backend:
            plt.pause(0.05)  # pragma: no cover (we are not using gui backend for testing)


class Monitor2DSpatial:
    r"""A Monitor for 2D steady-state problems
    """
    def __init__(self, check_on_x, check_on_y, check_every):
        self.using_non_gui_backend = matplotlib.get_backend() == 'agg'

        xy_tensor = torch.cartesian_prod(check_on_x, check_on_y)
        self.xx_tensor = torch.squeeze(xy_tensor[:, 0])
        self.yy_tensor = torch.squeeze(xy_tensor[:, 1])

        self.xx_array = self.xx_tensor.clone().detach().cpu().numpy()
        self.yy_array = self.yy_tensor.clone().detach().cpu().numpy()

        self.check_every = check_every

        self.fig = plt.figure(figsize=(30, 8))
        self.ax1 = self.fig.add_subplot(131)
        self.cb1 = None
        self.ax2 = self.fig.add_subplot(132)
        self.ax3 = self.fig.add_subplot(133)


    @staticmethod
    def _create_contour(ax, xx, yy, uu):
        triang = tri.Triangulation(xx, yy)
        contour = ax.tricontourf(triang, uu, cmap='coolwarm')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal', adjustable='box')
        return contour

    def check(self, approximator, history):
        self.ax1.clear()
        uu_array = approximator(self.xx_tensor, self.yy_tensor).detach().cpu().numpy()
        cs = self._create_contour(self.ax1, self.xx_array, self.yy_array, uu_array)
        if self.cb1 is None:
            self.cb1 = self.fig.colorbar(cs, format='%.0e', ax=self.ax1)
        else:
            self.cb1.mappable.set_clim(vmin=uu_array.min(), vmax=uu_array.max())
        self.ax1.set_title(f'approximation')

        self.ax2.clear()
        self.ax2.plot(history['train_loss'], label='training loss')
        self.ax2.plot(history['valid_loss'], label='validation loss')
        self.ax2.set_title('loss during training')
        self.ax2.set_ylabel('loss')
        self.ax2.set_xlabel('epochs')
        self.ax2.set_yscale('log')
        self.ax2.legend()

        self.ax3.clear()
        for metric_name, metric_values in history.items():
            if metric_name == 'train_loss' or metric_name == 'valid_loss':
                continue
            self.ax3.plot(metric_values, label=metric_name)
        self.ax3.set_title('metrics during training')
        self.ax3.set_ylabel('metrics')
        self.ax3.set_xlabel('epochs')
        self.ax3.set_yscale('log')

        # if there's not custom metrics, then there won't be any labels in this axis
        if len(history) > 2:
            self.ax3.legend()

        self.fig.canvas.draw()
        if not self.using_non_gui_backend:
            plt.pause(0.05)  # pragma: no cover (we are not using gui backend for testing)


def _solve_1dspatial_temporal(
    train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
    approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor
):
    r"""Solve a 1D time-dependent problem

    :param train_generator_spatial:
        A generator to generate 1D spatial points for training.
    :type train_generator_spatial: generator
    :param train_generator_temporal:
        A generator to generate 1D temporal points for training.
    :type train_generator_temporal: generator
    :param valid_generator_spatial:
        A generator to generate 1D spatial points for validation.
    :type valid_generator_spatial: generator
    :param valid_generator_temporal:
        A generator to generate 1D temporal points for validation.
    :type valid_generator_temporal: generator
    :param approximator:
        An approximator for 1D time-dependent problem.
    :type approximator: `temporal.SingleNetworkApproximator1DSpatialTemporal` or a custom `temporal.Approximator`
    :param optimizer:
        The optimization method to use for training.
    :type optimizer: `torch.optim.Optimizer`
    :param batch_size:
        The size of the mini-batch to use.
    :type batch_size: int
    :param max_epochs:
        The maximum number of epochs to train.
    :type max_epochs: int
    :param shuffle:
        Whether to shuffle the training examples every epoch.
    :type shuffle: bool
    :param metrics:
        Metrics to keep track of during training.
        The metrics should be passed as a dictionary where the keys are the names of the metrics,
        and the values are the corresponding function.
        The input functions should be the same as `pde` (of the approximator) and the output should be a numeric value.
        The metrics are evaluated on both the training set and validation set.
    :type metrics: dict[string, callable]
    :param monitor: The monitor to check the status of nerual network during training
    :type monitor: `temporal.Monitor1DSpatialTemporal` or `temporal.MonitorMinimal`
    """
    return _solve_spatial_temporal(
        train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
        approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor,
        train_routine=_train_1dspatial_temporal, valid_routine=_valid_1dspatial_temporal
    )


def _solve_2dspatial_temporal(
    train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
    approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor
):
    r"""Solve a 2D time-dependent problem

    :param train_generator_spatial:
        A generator to generate 2D spatial points for training.
    :type train_generator_spatial: generator
    :param train_generator_temporal:
        A generator to generate 1D temporal points for training.
    :type train_generator_temporal: generator
    :param valid_generator_spatial:
        A generator to generate 2D spatial points for validation.
    :type valid_generator_spatial: generator
    :param valid_generator_temporal:
        A generator to generate 1D temporal points for validation
    :type valid_generator_temporal: generator
    :param approximator:
        An approximator for 2D time-dependent problem.
    :type approximator: `temporal.SingleNetworkApproximator2DSpatialTemporal` or a custom `temporal.Approximator`
    :param optimizer:
        The optimization method to use for training.
    :type optimizer: `torch.optim.Optimizer`
    :param batch_size:
        The size of the mini-batch to use.
    :type batch_size: int
    :param max_epochs:
        The maximum number of epochs to train.
    :type max_epochs: int
    :param shuffle:
        Whether to shuffle the training examples every epoch.
    :type shuffle: bool
    :param metrics:
        Metrics to keep track of during training.
        The metrics should be passed as a dictionary where the keys are the names of the metrics,
        and the values are the corresponding function.
        The input functions should be the same as `pde` (of the approximator) and the output should be a numeric value.
        The metrics are evaluated on both the training set and validation set.
    :type metrics: dict[string, function]
    :param monitor:
        The monitor to check the status of nerual network during training.
    :type monitor: `temporal.Monitor2DSpatialTemporal` or `temporal.MonitorMinimal`
    """
    return _solve_spatial_temporal(
        train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
        approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor,
        train_routine=_train_2dspatial_temporal, valid_routine=_valid_2dspatial_temporal
    )

def _solve_2dspatial(
    train_generator_spatial, valid_generator_spatial,
    approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor
):
    r"""Solve a 2D steady-state problem

    :param train_generator_spatial:
        A generator to generate 2D spatial points for training.
    :type train_generator_spatial: generator
    :param valid_generator_spatial:
        A generator to generate 2D spatial points for validation.
    :type valid_generator_spatial: generator
    :param approximator:
        An approximator for 2D time-state problem.
    :type approximator: `temporal.SingleNetworkApproximator2DSpatial`, `temporal.SingleNetworkApproximator2DSpatialSystem`, or a custom `temporal.Approximator`
    :param optimizer:
        The optimization method to use for training.
    :type optimizer: `torch.optim.Optimizer`
    :param batch_size:
        The size of the mini-batch to use.
    :type batch_size: int
    :param max_epochs:
        The maximum number of epochs to train.
    :type max_epochs: int
    :param shuffle:
        Whether to shuffle the training examples every epoch.
    :type shuffle: bool
    :param metrics:
        Metrics to keep track of during training.
        The metrics should be passed as a dictionary where the keys are the names of the metrics,
        and the values are the corresponding function.
        The input functions should be the same as `pde` (of the approximator) and the output should be a numeric value.
        The metrics are evaluated on both the training set and validation set.
    :type metrics: dict[string, callable]
    :param monitor:
        The monitor to check the status of nerual network during training.
    :type monitor: `temporal.Monitor2DSpatial` or `temporal.MonitorMinimal`
    """
    return _solve_spatial_temporal(
        train_generator_spatial, None, valid_generator_spatial, None,
        approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor,
        train_routine=_train_2dspatial, valid_routine=_valid_2dspatial
    )


# _solve_1dspatial_temporal, _solve_2dspatial_temporal, _solve_2dspatial all call this function in the end
def _solve_spatial_temporal(
    train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
    approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor,
    train_routine, valid_routine
):
    history = {'train_loss': [], 'valid_loss': []}
    for metric_name, _ in metrics.items():
        history['train_' + metric_name] = []
        history['valid_' + metric_name] = []

    for epoch in range(max_epochs):
        train_epoch_loss, train_epoch_metrics = train_routine(
            train_generator_spatial, train_generator_temporal, approximator, optimizer, metrics, shuffle, batch_size
        )
        history['train_loss'].append(train_epoch_loss)
        for metric_name, metric_value in train_epoch_metrics.items():
            history['train_' + metric_name].append(metric_value)

        valid_epoch_loss, valid_epoch_metrics = valid_routine(
            valid_generator_spatial, valid_generator_temporal, approximator, metrics
        )
        history['valid_loss'].append(valid_epoch_loss)
        for metric_name, metric_value in valid_epoch_metrics.items():
            history['valid_' + metric_name].append(metric_value)

        if monitor and epoch % monitor.check_every == 0:
            monitor.check(approximator, history)

    return approximator, history


# training phase for 1D time-dependent problems
def _train_1dspatial_temporal(train_generator_spatial, train_generator_temporal,
                              approximator, optimizer, metrics, shuffle, batch_size):
    x = next(train_generator_spatial)
    t = next(train_generator_temporal)
    xx, tt = _cartesian_prod_dims(x, t)
    training_set_size = len(xx)
    idx = torch.randperm(training_set_size) if shuffle else torch.arange(training_set_size)

    batch_start, batch_end = 0, batch_size
    while batch_start < training_set_size:
        if batch_end > training_set_size:
            batch_end = training_set_size
        batch_idx = idx[batch_start:batch_end]
        batch_xx = xx[batch_idx]
        batch_tt = tt[batch_idx]

        batch_loss = approximator.calculate_loss(batch_xx, batch_tt, x, t)

        optimizer.zero_grad()
        batch_loss.backward()
        optimizer.step()

        batch_start += batch_size
        batch_end += batch_size

    epoch_loss = approximator.calculate_loss(xx, tt, x, t).item()

    epoch_metrics = approximator.calculate_metrics(xx, tt, x, t, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics


# training phase for 2D steady-state problems
def _train_2dspatial(train_generator_spatial, train_generator_temporal,
                     approximator, optimizer, metrics, shuffle, batch_size):
    xx, yy = next(train_generator_spatial)
    xx.requires_grad = True
    yy.requires_grad = True
    training_set_size = len(xx)
    idx = torch.randperm(training_set_size) if shuffle else torch.arange(training_set_size)

    batch_start, batch_end = 0, batch_size
    while batch_start < training_set_size:
        if batch_end > training_set_size:
            batch_end = training_set_size
        batch_idx = idx[batch_start:batch_end]
        batch_xx = xx[batch_idx]
        batch_yy = yy[batch_idx]

        batch_loss = approximator.calculate_loss(batch_xx, batch_yy)

        optimizer.zero_grad()
        batch_loss.backward()
        optimizer.step()

        batch_start += batch_size
        batch_end += batch_size

    epoch_loss = approximator.calculate_loss(xx, yy).item()

    epoch_metrics = approximator.calculate_metrics(xx, yy, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics


# validation phase for 2D steady-state problems
def _valid_2dspatial(valid_generator_spatial, valid_generator_temporal, approximator, metrics):
    xx, yy = next(valid_generator_spatial)
    xx.requires_grad = True
    yy.requires_grad = True

    epoch_loss = approximator.calculate_loss(xx, yy).item()

    epoch_metrics = approximator.calculate_metrics(xx, yy, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics


# training phase for 2D time-dependent problems
def _train_2dspatial_temporal(train_generator_spatial, train_generator_temporal,
                              approximator, optimizer, metrics, shuffle, batch_size):
    x, y = next(train_generator_spatial)
    t = next(train_generator_temporal)
    xx, tt = _cartesian_prod_dims(x, t)
    yy, tt = _cartesian_prod_dims(y, t)
    training_set_size = len(xx)
    idx = torch.randperm(training_set_size) if shuffle else torch.arange(training_set_size)

    batch_start, batch_end = 0, batch_size
    while batch_start < training_set_size:
        if batch_end > training_set_size:
            batch_end = training_set_size
        batch_idx = idx[batch_start:batch_end]
        batch_xx = xx[batch_idx]
        batch_yy = yy[batch_idx]
        batch_tt = tt[batch_idx]

        batch_loss = approximator.calculate_loss(batch_xx, batch_yy, batch_tt, x, y, t)

        optimizer.zero_grad()
        batch_loss.backward()
        optimizer.step()

        batch_start += batch_size
        batch_end += batch_size

    # TODO: this can give us the real loss after an epoch, but can be very memory intensive
    epoch_loss = approximator.calculate_loss(xx, yy, tt, x, y, t).item()

    epoch_metrics = approximator.calculate_metrics(xx, yy, tt, x, y, t, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics


# validation phase for 1D time-dependent problems
def _valid_1dspatial_temporal(valid_generator_spatial, valid_generator_temporal, approximator, metrics):
    x = next(valid_generator_spatial)
    t = next(valid_generator_temporal)
    xx, tt = _cartesian_prod_dims(x, t)

    epoch_loss = approximator.calculate_loss(xx, tt, x, t).item()

    epoch_metrics = approximator.calculate_metrics(xx, tt, x, t, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics


# validation phase for 2D time-dependent problems
def _valid_2dspatial_temporal(valid_generator_spatial, valid_generator_temporal, approximator, metrics):
    x, y = next(valid_generator_spatial)
    t = next(valid_generator_temporal)
    xx, tt = _cartesian_prod_dims(x, t)
    yy, tt = _cartesian_prod_dims(y, t)

    epoch_loss = approximator.calculate_loss(xx, yy, tt, x, y, t).item()

    epoch_metrics = approximator.calculate_metrics(xx, yy, tt, x, y, t, metrics)
    for k, v in epoch_metrics.items():
        epoch_metrics[k] = v.item()

    return epoch_loss, epoch_metrics



import pytest
import random
import torch
from neurodiffeq import diff
from neurodiffeq.networks import FCNN
from neurodiffeq.conditions import IVP, NoCondition
from neurodiffeq.generators import Generator1D
from neurodiffeq.solvers import GenericSolver, GenericSolution, BaseSolver
from neurodiffeq.solvers import Solver1D, Solver2D, SolverSpherical
from neurodiffeq.solvers import SolutionSphericalHarmonics

SPECIFIC_SOLVERS = [Solver1D, Solver2D, SolverSpherical]

N_SAMPLES = 64
T_MIN, T_MAX = 0.0, 1.0
DIFF_EQS = lambda u, t: [diff(u, t) + u]
CONDITIONS = [IVP(0, 1)]


@pytest.fixture
def generators():
    return dict(
        train=Generator1D(64, t_min=T_MIN, t_max=T_MAX, method='uniform'),
        valid=Generator1D(64, t_min=T_MIN, t_max=T_MAX, method='equally-spaced'),
    )


@pytest.fixture
def solver(generators):
    return GenericSolver(
        diff_eqs=DIFF_EQS,
        conditions=CONDITIONS,
        train_generator=generators['train'],
        valid_generator=generators['valid'],
        n_input_units=1,
        n_output_units=1,
    )


def test_legacies(solver, generators):
    solver.fit(1)

    assert solver.batch == solver._batch
    with pytest.warns(FutureWarning):
        assert solver._batch_examples == solver._batch

    with pytest.raises(TypeError), pytest.warns(FutureWarning):
        GenericSolver(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            train_generator=generators['train'],
            valid_generator=generators['valid'],
            criterion=lambda residuals, zeros: (residuals ** 2).mean(),
            n_input_units=1,
            n_output_units=1,
        ).fit(1)

    class SolverWithLegacyAdditionalLoss(BaseSolver):
        def additional_loss(self, funcs, key):
            return 0

    with pytest.raises(TypeError), pytest.warns(FutureWarning):
        SolverWithLegacyAdditionalLoss(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            train_generator=generators['train'],
            valid_generator=generators['valid'],
            n_input_units=1,
            n_output_units=1,
        ).fit(1)

    with pytest.warns(FutureWarning):
        GenericSolver(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            train_generator=generators['train'],
            valid_generator=generators['valid'],
            n_input_units=1,
            n_output_units=1,
            shuffle=True,
        )


def test_missing_generator(generators):
    with pytest.raises(ValueError):
        GenericSolver(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            train_generator=generators['train'],
            n_input_units=1,
            n_output_units=1,
        )
    with pytest.raises(ValueError):
        GenericSolver(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            valid_generator=generators['valid'],
            n_input_units=1,
            n_output_units=1,
        )
    with pytest.raises(ValueError):
        GenericSolver(
            diff_eqs=DIFF_EQS,
            conditions=CONDITIONS,
            n_input_units=1,
            n_output_units=1,
        )


def test_update_nonexistent_key_in_history(solver):
    with pytest.raises(KeyError):
        solver._update_history(1.0, metric_type='bad name', key='train')

    with pytest.raises(KeyError):
        solver._update_history(1.0, metric_type='bad name', key='valid')


@pytest.mark.parametrize(argnames='key', argvalues=['train', 'valid'])
def test_update_train_valid_history(solver, key):
    for _ in range(5):
        r = random.random()
        getattr(solver, f'_update_{key}_history')(value=r, metric_type='loss')
        assert solver.metrics_history[f'{key}_loss'][-1] == r


@pytest.mark.parametrize(argnames='key', argvalues=['train', 'valid'])
def test_generate_train_valid_batch(solver, key):
    batch = getattr(solver, f'_generate_{key}_batch')()
    assert batch == solver._batch[key]


def test_no_validation(solver):
    solver.n_batches['valid'] = 0
    solver.fit(1)


def test_lbfgs(generators):
    nets = [FCNN()]
    lbfgs = torch.optim.LBFGS(params=nets[0].parameters(), lr=1e-3)
    GenericSolver(
        diff_eqs=DIFF_EQS,
        conditions=CONDITIONS,
        train_generator=generators['train'],
        valid_generator=generators['valid'],
        nets=nets,
        optimizer=lbfgs,
        n_input_units=1,
        n_output_units=1,
    ).fit(1)


def test_early_stopping(solver):
    from neurodiffeq.callbacks import StopCallback

    cb = StopCallback()
    solver.fit(max_epochs=10, callbacks=[cb])
    assert solver.global_epoch == 1


def test_invalid_get_internals(solver):
    with pytest.raises(ValueError):
        solver.get_internals(['generator'], return_type='bad type')


def test_generic_solver(solver):
    solver.fit(1)


def test_generic_solution(solver):
    solution = solver.get_solution(best=False)
    t0 = torch.zeros((1, 1))
    assert (solution(t0) == 1).all()


@pytest.mark.parametrize(argnames='SolverClass', argvalues=SPECIFIC_SOLVERS)
def test_missing_domain(SolverClass, generators):
    with pytest.raises(ValueError):
        SolverClass(DIFF_EQS, CONDITIONS)
    with pytest.raises(ValueError):
        SolverClass(DIFF_EQS, CONDITIONS, train_generator=generators['train'])
    with pytest.raises(ValueError):
        SolverClass(DIFF_EQS, CONDITIONS, valid_generator=generators['valid'])


def test_default_generator():
    Solver1D(DIFF_EQS, CONDITIONS, t_min=0, t_max=1)
    Solver2D(DIFF_EQS, CONDITIONS, xy_min=(0, 0), xy_max=(1, 1))
    SolverSpherical(DIFF_EQS, CONDITIONS, r_min=0, r_max=1)


@pytest.mark.parametrize(argnames='SolverClass', argvalues=SPECIFIC_SOLVERS)
def test_get_internals_variables(SolverClass, generators):
    solver = SolverClass(
        DIFF_EQS,
        CONDITIONS,
        train_generator=generators['train'],
        valid_generator=generators['valid'],
    )
    d1 = BaseSolver._get_internal_variables(solver)
    d2 = solver._get_internal_variables()

    for k in d1:
        assert k in d2, f'{k} not in {d2.keys()}'
        assert d2[k] == d1[k], f'd1[{k}] == {d1[k]} != d2[{k}] == {d2[k]}'


def test_legacy_max_degree_in_solution_spherical_harmonics():
    with pytest.warns(FutureWarning):
        SolutionSphericalHarmonics(
            nets=[FCNN(1, 10)],
            conditions=[NoCondition()],
            max_degree=4,
        )
